import click
import keras
import pandas as pd
from pathlib import Path
from importlib import resources
from gresecml.core.session_manager import SessionManager
from gresecml.core.pcap_manager import lazy_read_packets
from gresecml.core.sniffer_manager import sniff_packets
from gresecml.core.progress_manager import ProgressManager
from gresecml.core.tensorflow_prediction_manager import tensorflow_predict

@click.group()
def tf():
    """Make predictions using TensorFlow."""
    pass

@click.command()
@click.option('--input', '-i', type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True, path_type=Path), help='Path to the .pcap file to analyze. If not provided, live capture will be used.')
@click.option('--output', '-o', type=click.Path(dir_okay=False, writable=True, path_type=Path), default=None, help='Path to save the output HTML file. Default is None.')
@click.option('--enable-full-output', '-efo', is_flag=True, help='Enable full output including all features in the HTML and console output.')
@click.option('--prop-normal-max', '-pnm', type=int, default=None, help='Set the maximum normal traffic probability threshold(%) for output. Default is None (no threshold).')
@click.option('--lazy-load', '-ll', is_flag=True, help='Enable lazy loading. Predicts sessions one by one to save memory.')
@click.option('--iface', '-if', type=str, default=None, help='Network interface for live capture. If not provided, the default interface will be used.')
@click.option('--timeout', '-t', type=int, default=60, help='Timeout in seconds for live capture. Default is 60 seconds.')
@click.option('--verbose', '-v', is_flag=True, help='Enable verbose output.')
def predict(input, output, enable_full_output, prop_normal_max, lazy_load, iface, timeout, verbose):
    """Make predictions for a live capture or a given pcap file using TensorFlow."""
    try:
        click.echo(click.style(f"\nAnalyzing network traffic using tensorflow model.\n", fg='yellow', bold=True))

        # Initialize session manager
        session_manager = SessionManager()

        # Get packets depending on pcap file or live capture
        if input is None:
            returned_sessions = session_manager.start_processing(sniff_packets(iface=iface, timeout=timeout))
        else:
            returned_sessions = session_manager.start_processing(lazy_read_packets(str(input)))

        if lazy_load:        
            # Lazy read packets and transform to sessions while making predictions
            progress_manager = ProgressManager()
            progress_manager.start(description="Processing network packets into sessions and making predictions")

            final_dataframe = pd.DataFrame()

            for session in returned_sessions:
                # Create input dictionary for prediction excluding non-feature columns
                input_dict = {col: session.iloc[0][col] for col in session.columns if col not in ['src_ip', 'dst_ip', 'src_port']}
                # Make predictions using TensorFlow model
                preds = tensorflow_predict(model=load_model(), input_data=input_dict)
                
                # Update session with predictions
                for key, value in preds.items():
                    session[key] = value
                
                # Save session if it meets the normal probability threshold
                if prop_normal_max is not None:
                    if session.iloc[0]['prediction_normal'] * 100 <= prop_normal_max:
                        final_dataframe = pd.concat([final_dataframe, session], ignore_index=True)
                else:
                    final_dataframe = pd.concat([final_dataframe, session], ignore_index=True)

            # Output results
            dataframe_to_html(final_dataframe, output, enable_full_output)

            if verbose:
                print_dataframe(final_dataframe, enable_full_output)
                    
            progress_manager.stop()
            
        if not lazy_load:
            # For predicting all sessions together after processing (stores all sessions in memory)
            # Load packets and transform to sessions
            progress_manager = ProgressManager()
            progress_manager.start(description="Processing network packets into sessions")

            sessions = pd.DataFrame()

            for session in returned_sessions:
                sessions = pd.concat([sessions, session], ignore_index=True)

            progress_manager.stop()
        
            # Make predictions using TensorFlow model
            progress_manager = ProgressManager(True)
            progress_manager.start(description="Making predictions", total=len(sessions))

            columns = sessions.columns.tolist()

            for idx, session in sessions.iterrows():
                input_dict = {col: session[col] for col in columns if col not in ['src_ip', 'dst_ip', 'src_port']}

                preds = tensorflow_predict(model=load_model(), input_data=input_dict)
                
                for key, value in preds.items():
                    sessions.loc[idx, key] = value
                
                progress_manager.advance()

            progress_manager.stop()

            # Display predictions
            if prop_normal_max is not None:
                filtered_sessions: pd.DataFrame = sessions[sessions['prediction_normal'] * 100 <= prop_normal_max]
                dataframe_to_html(filtered_sessions, output, enable_full_output)
                # Verbose output to console
                if verbose:
                    print_dataframe(filtered_sessions, enable_full_output)
            else:
                dataframe_to_html(sessions, output, enable_full_output)
                # Verbose output to console
                if verbose:
                    print_dataframe(sessions, enable_full_output)

        click.echo(click.style("\nPrediction process completed.\n", fg='yellow', bold=True))
    except Exception as e:
        click.echo(click.style(f"\nAn error occurred during prediction: \"{str(e)}\"\n", fg='red', bold=True))

tf.add_command(predict)

# Load TensorFlow model from package resources
def load_model():
    """Load TensorFlow model from package resources."""
    # Use with to get a temporary path to the model file
    with resources.path('gresecml.ml_models', 'tensorflow.keras') as model_path:
        model = keras.models.load_model(model_path)
    return model

# Convert DataFrame to HTML file with or without full output
def dataframe_to_html(df: pd.DataFrame, path: Path = None, enable_full_output: bool = False):
    """Convert DataFrame to HTML file."""


    # Drop unwanted columns if they exist
    unwanted_columns = ['time_mean', 'time_std', 'bytes_mean', 'bytes_std', 'inter_arrival_std', 'inter_arrival_min', 'inter_arrival_max']

    if not enable_full_output:
        for col in unwanted_columns:
            if col in df.columns:
                df = df.drop(columns=[col])

    if path is not None:
        output_path = Path.cwd() / path
        click.echo(click.style(f"Saving output to: {output_path}", fg='yellow'))
        df.to_html(output_path)

# Print DataFrame to console with or without full output
def print_dataframe(df: pd.DataFrame, enable_full_output: bool = False):
    """Print DataFrame to console."""

    # Drop unwanted columns if they exist
    unwanted_columns = ['time_mean', 'time_std', 'bytes_mean', 'bytes_std', 'inter_arrival_std', 'inter_arrival_min', 'inter_arrival_max', 'flag_syn_count', 'flag_ack_count', 'flag_fin_count', 'flag_rst_count', 'flag_dns_count', 'successful_connection', 'num_of_dest_unreachable']

    if not enable_full_output:
        for col in unwanted_columns:
            if col in df.columns:
                df = df.drop(columns=[col])

    click.echo(df.to_string(index=False))


import tensorflow as tf
import keras

def tensorflow_predict(model: keras.Model, input_data: dict) -> dict[str, float]:
    """Make predictions using a TensorFlow model and input data.
    
    Args:
        model (keras.Model): The TensorFlow model to use for predictions.
        input_data (dict): Input data for making predictions.
    
    Returns:
        dict[str, float]: Dictionary of predictions made by the model. prediction_normal and prediction_portscan
    """
    # Convert input data to tensors
    input_dict = {name: tf.convert_to_tensor([value]) for name, value in input_data.items()}

    # Make predictions
    try:
        preds = model(input_dict)
    except Exception:
        raise RuntimeError("Failed to make predictions with the TensorFlow model.")
    
    try:
        # Update output dictionary if more categories are added in the future
        output_dict = {
            "prediction_normal": float(preds[0][0]),
            "prediction_portscan": float(preds[0][1]),
        }
    except Exception:
        raise RuntimeError("Failed to process predictions from the TensorFlow model.")

    return output_dict


from scapy.all import Packet, AsyncSniffer
from queue import Empty, Queue
from typing import Iterator

def sniff_packets(iface: str = None, timeout: float = 10) -> Iterator[Packet]:
        queue = Queue()

        sniffer = AsyncSniffer(iface=iface, timeout=timeout, store=True, prn=lambda x: queue.put_nowait(x))
        sniffer.start()

        while sniffer.running:
            try:
                yield queue.get(block=False)
            except Empty:
                # no packet available yet, loop again until overall timeout
                continue

        # Wait for the sniffer to finish
        sniffer.join()

        # Get any remaining packets from the queue
        while not queue.empty():
            yield queue.get(block=False)

        # Stop the sniffer if it's still running
        try:
            sniffer.stop()
        except Exception:
            pass


import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' # Suppress TensorFlow logging from console
import click
from gresecml.commands.tensorflow_group import tf as tf_group


@click.group()
def main():
    pass
main.add_command(tf_group)


from rich.progress import Progress, BarColumn, TextColumn, TimeElapsedColumn

class ProgressManager:
    def __init__(self, with_pct: bool = False) -> None:
        # Configure the progress bar layout
        if with_pct:
            self.progress = Progress(
                TextColumn("[bold blue]{task.description}"),
                BarColumn(),
                "[progress.percentage]{task.percentage:>3.0f}%",
                TimeElapsedColumn(),
            )
        else:
            self.progress = Progress(
                TextColumn("[bold blue]{task.description}"),
                BarColumn(),
                TimeElapsedColumn(),
            )
        self.task_id = None

    def start(self, description: str, total: int = None) -> None:
        """Start the progress display and add a task."""
        self.progress.start()
        self.task_id = self.progress.add_task(description, total=total)

    def remove_task(self) -> None:
        """Remove a task from the progress display."""
        self.progress.remove_task(self.task_id)

    def advance(self, step: int = 1) -> None:
        """Advance the progress by a given step."""
        self.progress.update(self.task_id, advance=step)

    def stop(self) -> None:
        """Stop the progress display."""
        self.progress.stop()

    def end(self) -> None:
        """Stop the progress display and clear all tasks."""
        self.remove_task()
        self.progress.stop()
        self.progress = None
        self.task_id = None




import time
from gresecml.models.gre_packet import GrePacket

class GreSession:

    gre_packets: list[GrePacket]

    def __init__(self):
        self.gre_packets: list[GrePacket] = []
        self.lastest_arrival: float = 0.0

    def add_packet(self, packet: GrePacket):
        # Add the packet to the session
        self.gre_packets.append(packet)
        # Update latest arrival time for session timeout tracking
        self.lastest_arrival = time.time()

    @property
    def session_sid(self) -> tuple:
        if len(self.gre_packets) > 0:
            return self.gre_packets[0].sid
        else:
            return tuple()

    @property
    def packets(self) -> list[GrePacket]:
        return self.gre_packets
    
    @property
    def successful_connection(self) -> bool:
        # Check if session contains a successful TCP handshake (SYN, SYN-ACK, ACK)
        if len(self.gre_packets) > 0:
            sorted_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)

        if sorted_packets[0].protocol == "TCP":
            syn_seen = False
            syn_ack_seen = False
            ack_seen = False

            for packet in sorted_packets:
                if packet.flag == "SYN":
                    syn_seen = True
                elif packet.flag == "SYN-ACK" and syn_seen:
                    syn_ack_seen = True
                elif packet.flag == "ACK" and syn_ack_seen:
                    ack_seen = True
                    break

            return syn_seen and syn_ack_seen and ack_seen
        elif sorted_packets[0].protocol == "UDP":
            if self.num_of_dest_unreachable == 0:
                return True
            else:
                return False
        elif sorted_packets[0].protocol == "ICMP":
            return True
        return False
    
    @property
    def num_of_dest_unreachable(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.is_unreachable_dest)

# IPs
    @property
    def first_src_ip(self) -> str:
        if len(self.gre_packets) > 0:
            return self.gre_packets[0].src_ip
        return ""

    @property
    def first_dst_ip(self) -> str:
        if len(self.gre_packets) > 0:
            return self.gre_packets[0].dst_ip
        return ""

# Ports
    @property
    def first_src_port(self) -> int:
        if len(self.gre_packets) > 0:
            return self.gre_packets[0].src_port
        return 0
    
    @property
    def first_dst_port(self) -> int:
        if len(self.gre_packets) > 0:
            return self.gre_packets[0].dst_port
        return 0

# Protocols
    @property
    def first_protocol(self) -> str:
        if len(self.gre_packets) > 0:
            return self.gre_packets[0].protocol
        return "None"
    
    @property
    def protocol_tcp_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.protocol == "TCP")

    @property
    def protocol_udp_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.protocol == "UDP")

    @property
    def protocol_icmp_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.protocol == "ICMP")

    @property
    def protocol_other_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.protocol not in ["TCP", "UDP", "ICMP"])

# Flags
    @property
    def flag_syn_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "SYN")

    @property
    def flag_syn_ack_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "SYN-ACK")

    @property
    def flag_psh_ack_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "PSH-ACK")

    @property
    def flag_fin_ack_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "FIN-ACK")

    @property
    def flag_rst_ack_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "RST-ACK")

    @property
    def flag_rst_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "RST")

    @property
    def flag_fin_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "FIN")

    @property
    def flag_ack_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "ACK")

    @property
    def flag_psh_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "PSH")

    @property
    def flag_urg_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "URG")

    @property
    def flag_other_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.flag == "OTHER")

# Number of DNS Queries
    @property
    def dns_query_count(self) -> int:
        return sum(1 for packet in self.gre_packets if packet.is_dns)

# Time
    @property
    def total_time(self) -> float:
        if len(self.gre_packets) > 1:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            return ordered_packets[-1].scapy_packet.time - ordered_packets[0].scapy_packet.time
        return 0.0
    
    @property
    def mean_time(self) -> float:
        if len(self.gre_packets) > 1:
            return self.total_time / len(self.gre_packets)
        return 0.0
    
    @property
    def std_time(self) -> float:
        if len(self.gre_packets) > 1:
            mean = self.mean_time
            variance = sum((packet.scapy_packet.time - mean) ** 2 for packet in self.gre_packets) / len(self.gre_packets)
            return variance ** 0.5
        return 0.0
    
    @property
    def first_packet_time(self) -> float:
        if len(self.gre_packets) > 0:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            return ordered_packets[0].scapy_packet.time
        return 0.0
    
    @property
    def last_packet_time(self) -> float:
        if len(self.gre_packets) > 0:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            return ordered_packets[-1].scapy_packet.time
        return 0.0

# Flow
    @property
    def flow_pr_sec(self) -> float:
        if len(self.gre_packets) > 0:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            total_bytes = sum(packet.length for packet in ordered_packets)
            total_time = ordered_packets[-1].scapy_packet.time - ordered_packets[0].scapy_packet.time
            if total_time > 0:
                return total_bytes / total_time
        return 0.0
    
    @property
    def fwd_flow_pr_sec(self) -> float:
        if len(self.gre_packets) > 0:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            total_bytes = sum(packet.length for packet in ordered_packets if packet.src_ip == self.first_src_ip)
            total_time = ordered_packets[-1].scapy_packet.time - ordered_packets[0].scapy_packet.time
            if total_time > 0:
                return total_bytes / total_time
        return 0.0

    @property
    def bwd_flow_pr_sec(self) -> float:
        if len(self.gre_packets) > 0:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            total_bytes = sum(packet.length for packet in ordered_packets if packet.src_ip == self.first_dst_ip)
            total_time = ordered_packets[-1].scapy_packet.time - ordered_packets[0].scapy_packet.time
            if total_time > 0:
                return total_bytes / total_time
        return 0.0

# Lengths
    @property
    def length(self) -> int:
        if len(self.gre_packets) > 0:
            return sum(packet.length for packet in self.gre_packets)
        return 0
    
    @property
    def fwd_length(self) -> int:
        if len(self.gre_packets) > 0:
            return sum(packet.length for packet in self.gre_packets if packet.src_ip == self.first_src_ip)
        return 0
    
    @property
    def bwd_length(self) -> int:
        if len(self.gre_packets) > 0:
            return sum(packet.length for packet in self.gre_packets if packet.src_ip == self.first_dst_ip)
        return 0
    
    @property
    def length_mean(self) -> float:
        if len(self.gre_packets) > 0:
            #Gennemsnit af længder
            return sum(packet.length for packet in self.gre_packets) / len(self.gre_packets)
        return 0.0

    @property
    def length_min(self) -> float:
        if len(self.gre_packets) > 0:
            return min(packet.length for packet in self.gre_packets)
        return 0.0

    @property
    def length_max(self) -> float:
        if len(self.gre_packets) > 0:
            return max(packet.length for packet in self.gre_packets)
        return 0.0
    
    @property
    def length_std(self) -> float:
        if len(self.gre_packets) > 0:
            # Gennemsnit
            mean = self.length_mean
            # Finder varians, gennemsnittet af de kvadrerede afvigelser fra middelværdien
            variance = sum((packet.length - mean) ** 2 for packet in self.gre_packets) / len(self.gre_packets)
            # Kvadratroden af variansen for at finde standardafvigelsen
            return variance ** 0.5
        return 0.0

# Packet counts
    @property
    def packet_count(self) -> int:
        return len(self.gre_packets)
    
# Inter arrival times
    @property
    def inter_arrival_times(self) -> list[float]:
        inter_arrivals = []
        if len(self.gre_packets) > 1:
            ordered_packets = sorted(self.gre_packets, key=lambda p: p.scapy_packet.time)
            for i in range(1, len(ordered_packets)):
                inter_arrival = ordered_packets[i].scapy_packet.time - ordered_packets[i-1].scapy_packet.time
                inter_arrivals.append(inter_arrival)
        return inter_arrivals


    @property
    def inter_arrival_mean(self) -> float:
        inter_arrivals = self.inter_arrival_times
        if len(inter_arrivals) > 0:
            return sum(inter_arrivals) / len(inter_arrivals)
        return 0.0
    
    @property
    def inter_arrival_min(self) -> float:
        inter_arrivals = self.inter_arrival_times
        if len(inter_arrivals) > 0:
            return min(inter_arrivals)
        return 0.0
    
    @property
    def inter_arrival_max(self) -> float:
        inter_arrivals = self.inter_arrival_times
        if len(inter_arrivals) > 0:
            return max(inter_arrivals)
        return 0.0
    
    @property
    def inter_arrival_variance(self) -> float:
        inter_arrivals = self.inter_arrival_times
        if len(inter_arrivals) > 0:
            mean = sum(inter_arrivals) / len(inter_arrivals)
            variance = sum((iat - mean) ** 2 for iat in inter_arrivals) / len(inter_arrivals)
            return variance
        return 0.0
    
    @property
    def inter_arrival_std(self) -> float:
        return self.inter_arrival_variance ** 0.5

    




import time
import pandas as pd
from typing import Iterator
from scapy.all import TCP, UDP, ICMP, IP, Packet
from gresecml.models.gre_packet import GrePacket
from gresecml.models.gre_session import GreSession
import gresecml.core.pcap_manager as pcap_manager
import gresecml.core.sniffer_manager as sniffer_manager

class SessionManager:
    def __init__(self, tcp_timeout=300, tcp_rst_fin_timeout=20, udp_timeout=60.0, icmp_timeout=60.0):
        self.sessions: dict[tuple, GreSession] = {}
        self.open_sessions: dict[tuple, float] = {}
        self.closing_sessions: dict[tuple, float] = {}
        self.tcp_timeout: float = tcp_timeout # Timeout for established TCP sessions
        self.tcp_rst_fin_timeout: float = tcp_rst_fin_timeout # Timeout for closing TCP sessions (FIN, RST)
        self.udp_timeout: float = udp_timeout # Timeout for inactive UDP sessions
        self.icmp_timeout: float = icmp_timeout # Timeout for inactive ICMP sessions
        self.dataframes_for_export: list[pd.DataFrame] = []

    def start_processing(self, packets: Iterator[Packet]) -> Iterator[pd.DataFrame]:
        self.sessions.clear()
        self.open_sessions.clear()
        self.closing_sessions.clear()

        for packet in packets:
            gre_packet = GrePacket(packet)
            # Håndter pakken baseret på dens protokol
            try:
                self.handle_gre_packet(gre_packet)
            except Exception as e:
                # Burde logge fejl her
                continue

            # Eksporter dataframes hvis der er nogen
            for df in self.export_dataframes():
                yield df

        # Eksporter alle resterende sessioner efter pcap er færdig
        for sid in list(self.sessions.keys()):
            try:
                self.add_to_export(sid)
            except Exception as e:
                # Burde logge fejl her
                continue

        # Eksporter sidste dataframes hvis der er nogen
        for df in self.export_dataframes():
            yield df

#------------------------------------------------------------------------------------------------------------------------------------------------
# ---- Hjælpe metoder ----
#------------------------------------------------------------------------------------------------------------------------------------------------
    def handle_gre_packet(self, gre_packet: GrePacket) -> None:
        if not gre_packet.scapy_packet.haslayer(IP):
            # Ingen IPv6 support
            return
        if gre_packet.scapy_packet.haslayer(TCP):
            self.handle_tcp(gre_packet)
        elif gre_packet.scapy_packet.haslayer(UDP):
            self.handle_udp(gre_packet)
        elif gre_packet.scapy_packet.haslayer(ICMP): 
            self.handle_icmp(gre_packet)
        else:
            return

    # Starter en ny session
    def start_session(self, packet: GrePacket) -> None:
        self.sessions[packet.sid] = GreSession()
        self.add_packet_to_session(packet)
        self.mark_open_session(packet.sid)

    # Tilføjer en pakke til en eksisterende session
    def add_packet_to_session(self, packet: GrePacket) -> None:
        self.sessions[packet.sid].add_packet(packet)

    # Marker en session som åben - til at fange første dele af tcp forbindelsen ved SYN
    def mark_open_session(self, sid) -> None:
        self.open_sessions[sid] = time.time()

    # Marker en session som lukker - til at fange sidste dele af tcp forbindelsen ved FIN-ACK
    def mark_closing_session(self, sid) -> None:
        if sid in self.open_sessions:
            del self.open_sessions[sid]
        self.closing_sessions[sid] = time.time()

#------------------------------------------------------------------------------------------------------------------------------------------------
# ---- Håndtering af pakker efter protokol ----
#------------------------------------------------------------------------------------------------------------------------------------------------

    def handle_tcp(self, gre_packet: GrePacket) -> None:
        # Tjek om timeouts er overskredet for åbne sessions
        for sid in self.open_sessions.copy():
            if self.open_sessions[sid] + self.tcp_timeout < time.time():
                self.mark_closing_session(sid)

        # Tjek om timeouts er overskredet for closing sessions
        for sid in self.closing_sessions.copy():
            if self.closing_sessions[sid] + self.tcp_rst_fin_timeout < time.time():
                self.add_to_export(sid)

        # Håndter SYN pakker - start ny session eller eksporter eksisterende før ny session startes
        if gre_packet.flag == "SYN" and gre_packet.sid not in self.sessions:
            self.start_session(gre_packet)
            return
        elif gre_packet.flag == "SYN" and gre_packet.sid in self.sessions:
            self.add_to_export(gre_packet.sid)
            self.start_session(gre_packet)
            return
        
        # Hvis sessionen eksisterer, tilføj pakken
        if gre_packet.sid in self.sessions:
            self.add_packet_to_session(gre_packet)
            if gre_packet.flag == "FIN":
                self.mark_closing_session(gre_packet.sid)
            elif gre_packet.flag == "RST":
                self.mark_closing_session(gre_packet.sid)
        #Hvis ikke-syn sessions også skal med
        else:
            self.start_session(gre_packet)
            
    def handle_udp(self, gre_packet: GrePacket) -> None:
        if gre_packet.sid in self.sessions:
            if gre_packet.scapy_packet.time - self.sessions[gre_packet.sid].first_packet_time < self.udp_timeout:
                self.add_packet_to_session(gre_packet)
            else:
                self.add_to_export(gre_packet.sid)
                self.start_session(gre_packet)
        else:
            self.start_session(gre_packet)

    def handle_icmp(self, gre_packet: GrePacket) -> None:
        if gre_packet.sid in self.sessions:
            if gre_packet.scapy_packet.time - self.sessions[gre_packet.sid].first_packet_time < self.icmp_timeout:
                self.add_packet_to_session(gre_packet)
            else:
                self.add_to_export(gre_packet.sid)
                self.start_session(gre_packet)
        else:
            self.start_session(gre_packet)

#------------------------------------------------------------------------------------------------------------------------------------------------
# ---- Eksport af sessioner ----
#------------------------------------------------------------------------------------------------------------------------------------------------

    def add_to_export(self, session_id: str) -> None:
        if session_id in self.sessions:
            # Fjern sessionen fra aktiv liste
            session = self.sessions.pop(session_id)
            
            # Fjern fra åbne sessions hvis den er der
            if session_id in self.open_sessions:
                del self.open_sessions[session_id]
            
            # Fjern fra closing sessions hvis den er der
            if session_id in self.closing_sessions:
                del self.closing_sessions[session_id]
            
            # Eksporter session data som dataframe
            data = {
                "src_ip": (str)(session.first_src_ip),
                "dst_ip": (str)(session.first_dst_ip),
                "src_port": (str)(session.first_src_port),
                "dst_port": (str)(session.first_dst_port),
                "packets_total": (int)(session.packet_count),
                "time_total": (float)(session.total_time),
                "time_mean": (float)(session.mean_time),
                "time_std": (float)(session.std_time),
                "bytes_total": (int)(session.length),
                "bytes_mean": (float)(session.length_mean),
                "bytes_std": (float)(session.length_std),
                "bytes_fwd": (int)(session.fwd_length),
                "bytes_bwd": (int)(session.bwd_length),
                "flow_pr_sec": (float)(session.flow_pr_sec),
                "inter_arrival_mean": (float)(session.inter_arrival_mean),
                "inter_arrival_std": (float)(session.inter_arrival_std),
                "inter_arrival_min": (float)(session.inter_arrival_min),
                "inter_arrival_max": (float)(session.inter_arrival_max),
                "protocol": (str)(session.first_protocol),
                "flag_syn_count": (int)(session.flag_syn_count),
                "flag_ack_count": (int)(session.flag_ack_count),
                "flag_fin_count": (int)(session.flag_fin_count),
                "flag_rst_count": (int)(session.flag_rst_count),
                "flag_dns_count": (int)(session.dns_query_count),
                "successful_connection": (str)(session.successful_connection),
                "num_of_dest_unreachable": (int)(session.num_of_dest_unreachable)
            }
            # Tilføj data til eksport listen
            self.dataframes_for_export.append(pd.DataFrame([data]))


    def export_dataframes(self) -> Iterator[pd.DataFrame]:
        for session in self.dataframes_for_export:
            yield session
        self.dataframes_for_export.clear()
    



from scapy.all import PcapReader, rdpcap, Packet

def read_packets(pcap_file_path: str) -> list[Packet]:
        try:
            packets = rdpcap(pcap_file_path)
            return packets
        except Exception as e:
            raise ValueError("Not a valid pcap file or unable to read the file.")
        
def lazy_read_packets(pcap_file_path: str):
        try:
            with PcapReader(pcap_file_path) as pcap_reader:
                for packet in pcap_reader:
                    yield packet
        except Exception as e:
            raise ValueError("Not a valid pcap file or unable to read the file.")



from scapy.packet import Packet as Scapy_packet
from scapy.layers.inet import IP, TCP, UDP, ICMP
from scapy.layers.dns import DNS
from scapy.all import *

class GrePacket:
    def __init__(self, scapy_packet: Scapy_packet):
        self.sid = None
        self.scapy_packet: Scapy_packet = scapy_packet

        self.set_sid()

    def set_sid(self):
        if self.sid == None:
            sorted_ips: List[str] = sorted([self.src_ip, self.dst_ip])
            sorted_ports: List[str] = sorted([self.src_port, self.dst_port])

            self.sid = (sorted_ips[0], sorted_ips[1], sorted_ports[0], sorted_ports[1])

    @property
    def src_ip(self) -> str:
        if self.scapy_packet.haslayer(IP):
            return self.scapy_packet[IP].src
        return "0.0.0.0"

    @property
    def dst_ip(self) -> str:
        if self.scapy_packet.haslayer(IP):
            return self.scapy_packet[IP].dst
        return "0.0.0.0"

    @property
    def src_port(self) -> str:
        if self.scapy_packet.haslayer(TCP):
            return (str)(self.scapy_packet[TCP].sport)
        elif self.scapy_packet.haslayer(ICMP) and self.scapy_packet[ICMP].type == 3:
            inner = IP(bytes(self.scapy_packet[ICMP].payload))
            return (str)(inner[UDP].sport) if inner.haslayer(UDP) else "0"
        elif self.scapy_packet.haslayer(UDP):
            return (str)(self.scapy_packet[UDP].sport)
        return "0"

    @property
    def dst_port(self) -> str:
        if self.scapy_packet.haslayer(TCP):
            return (str)(self.scapy_packet[TCP].dport)
        elif self.scapy_packet.haslayer(ICMP) and self.scapy_packet[ICMP].type == 3:
            inner = IP(bytes(self.scapy_packet[ICMP].payload))
            return (str)(inner[UDP].dport) if inner.haslayer(UDP) else "0"
        elif self.scapy_packet.haslayer(UDP):
            return (str)(self.scapy_packet[UDP].dport)
        return "0"
    
    @property
    def is_unreachable_dest(self) -> bool:
        return self.scapy_packet.haslayer(ICMP) and self.scapy_packet[ICMP].type == 3

    @property
    def protocol(self) -> str:
        if self.scapy_packet.haslayer(IP):
            if self.scapy_packet.haslayer(UDP):
                return "UDP"
            elif self.scapy_packet.haslayer(TCP):
                return "TCP"
            elif self.scapy_packet.haslayer(ICMP):
                return "ICMP"
            return "Other"
        return "Other"

    @property
    def flag(self) -> str:
        if self.scapy_packet.haslayer(TCP):
            flags = self.scapy_packet[TCP].flags

            if flags & 0x02 and flags & 0x10:
                return "SYN-ACK"
            elif flags & 0x10 and flags & 0x08:
                return "PSH-ACK"
            elif flags & 0x01 and flags & 0x10:
                return "FIN-ACK"
            elif flags & 0x04 and flags & 0x10:
                return "RST-ACK"
            elif flags == 0x02:
                return "SYN"
            elif flags == 0x10:
                return "ACK"
            elif flags == 0x04:
                return "RST"
            elif flags == 0x01:
                return "FIN"
            elif flags == 0x08:
                return "PSH"
            elif flags == 0x20:
                return "URG"
            else:
                return "OTHER"
        return "None"
    
    @property
    def is_dns(self) -> bool:
        return self.scapy_packet.haslayer(DNS)

    @property
    def length(self) -> int:
        return len(self.scapy_packet)

    @property
    def window_size(self) -> int:
        if self.scapy_packet.haslayer(TCP):
            return self.scapy_packet[TCP].window
        return 0
    
    @property
    def inter_arrival(self) -> float | None:
        if self.prev is not None:
            return self.scapy_packet.time - self.prev.scapy_packet.time
        return None